fn main() {
    println!("Hello, world!");
}

#[cfg(test)]
#[allow(dead_code)]
mod unnested {
    use rustructure::{Walkable, StringWalker};

    #[derive(Walkable)]
    struct EmptyStruct {}

    #[derive(Walkable)]
    struct SingletonStruct {
        value: u8
    }

    #[derive(Walkable)]
    struct DualStruct {
        value: u8,
        second_value: u8,
    }

    #[derive(Walkable)]
    struct SingletonNewtypeStruct(u8);

    #[derive(Walkable)]
    struct DualNewtypeStruct(u8, u8);

    #[derive(Walkable)]
    enum UnitEnum {}

    #[derive(Walkable)]
    enum SingleVariantEnum {
        A
    }

    #[derive(Walkable)]
    enum DoubleVariantEnum {
        A,
        B
    }

    #[test]
    fn struct_checks() {
        assert_eq!(
            StringWalker::walk::<EmptyStruct>(),
            "(struct:EmptyStruct)".to_owned()
        );

        assert_eq!(
            StringWalker::walk::<SingletonStruct>(),
            "(struct:SingletonStruct (field:value integer))".to_owned()
        );

        assert_eq!(
            StringWalker::walk::<DualStruct>(),
            "(struct:DualStruct (field:value integer) (field:second_value integer))".to_owned()
        );

        assert_eq!(
            StringWalker::walk::<SingletonNewtypeStruct>(),
            "(struct:SingletonNewtypeStruct (field:0 integer))".to_owned()
        );

        assert_eq!(
            StringWalker::walk::<DualNewtypeStruct>(),
            "(struct:DualNewtypeStruct (field:0 integer) (field:1 integer))".to_owned()
        );
    }

    #[test]
    fn enum_checks() {
        assert_eq!(
            StringWalker::walk::<UnitEnum>(),
            "(enum:UnitEnum)".to_owned()
        );

        assert_eq!(
            StringWalker::walk::<SingleVariantEnum>(),
            "(enum:SingleVariantEnum (variant:A))".to_owned()
        );

        assert_eq!(
            StringWalker::walk::<DoubleVariantEnum>(),
            "(enum:DoubleVariantEnum (variant:A) (variant:B))".to_owned()
        );
    }
}

#[cfg(test)]
#[allow(dead_code)]
mod simple_nested {
    use rustructure::{Walkable, StringWalker};

    #[derive(Walkable)]
    struct Contained {
        v: u8
    }

    #[derive(Walkable)]
    struct Container {
        c: Contained
    }

    #[test]
    fn simple_container() {
        assert_eq!(
            StringWalker::walk::<Contained>(),
            "(struct:Contained (field:v integer))".to_owned()
        );

        assert_eq!(
            StringWalker::walk::<Container>(),
            "(struct:Container (field:c (struct:Contained (field:v integer))))".to_owned()
        );
    }

    #[derive(Walkable)]
    enum IntegerHolder {
        Hold8(u8)
    }

    #[derive(Walkable)]
    enum StructHolder {
        HoldContained(Contained)
    }

    #[test]
    fn nonunit_variant() {
        assert_eq!(
            StringWalker::walk::<IntegerHolder>(),
            "(enum:IntegerHolder (variant:Hold8 (field:0 integer)))".to_owned()
        );

        assert_eq!(
            StringWalker::walk::<StructHolder>(),
            "(enum:StructHolder (variant:HoldContained (field:0 (struct:Contained (field:v integer)))))".to_owned()
        );
    }

    #[test]
    fn simple_tuples() {
        assert_eq!(
            StringWalker::walk::<(u8,)>(),
            "(tuple (field:0 integer))".to_owned()
        );

        assert_eq!(
            StringWalker::walk::<(u8,u8)>(),
            "(tuple (field:0 integer) (field:1 integer))".to_owned()
        );

        assert_eq!(
            StringWalker::walk::<(u8,u8,u8)>(),
            "(tuple (field:0 integer) (field:1 integer) (field:2 integer))".to_owned()
        );

        assert_eq!(
            StringWalker::walk::<(u8,u8,u8,u8)>(),
            "(tuple (field:0 integer) (field:1 integer) (field:2 integer) (field:3 integer))".to_owned()
        );

        assert_eq!(
            StringWalker::walk::<((u8,),)>(),
            "(tuple (field:0 (tuple (field:0 integer))))".to_owned()
        );
    }
}
